#include "computation/computation_api.h"
#include "computation/backup/backup_api.h"
#include "computation/backup/dijkstra.h"
#include "database/database_api.h"
#include "common/common.h"

#include <stdio.h>
#include <string.h>

#if defined LINUX && defined TIMING
#include <sys/time.h>
#endif

// Function for backup path calculation
double computeCost(DataBase* dataBase, DBLinkState** lsList, long src, long dst, 
		   CPDijkNode* dn, DBLabelSwitchedPath* lsp, BackupType type)
{
    CPDijkNode* ptr;
    LSPRequest newReq;
    DBLinkState newLS;
    long linkID,i,j;
    double alpha, beta;

    double inc;
    double bw_before[NB_OA];
    double bw_after[NB_OA];
    double bw_tot_bef=0, bw_tot_aft=0;
    double bw_tot=0;

    if (DBlinkStateInit(&newLS) < 0)
    {
	addError(CRITICAL,"Unable to init a link state in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    beta = 1.0/(1+10);
    alpha = 1 - beta;

    // bandwidth increment
    // -------------------

    lspRequestInit(&newReq);
    newReq.primID = lsp->id;
    
    if (lsp->precedence + 1 < NB_PREEMPTION)
	newReq.precedence = lsp->precedence + 1;
    else
	newReq.precedence = lsp->precedence;

    memmove(&(newReq.bw), &(lsp->bw), NB_OA * sizeof(double));

    if (type == LOCAL)
	newReq.type = LOCAL_BACK;
    else if (type == GLOBAL)
	newReq.type = GLOBAL_BACK;
    else
    {
	addError(WARNING,"Unknown backup type in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    longListPushBack(&(newReq.path), dst);

    ptr = dn;
    while (ptr != NULL)
    {
	longListPushBack(&(newReq.path), ptr->node->id);
	ptr = ptr->from;
    }

    // now reverse the path
    for (i=0; i<(newReq.path.top + 1)/2; i++)
    {
	int tmp = newReq.path.cont[i];
	newReq.path.cont[i] = newReq.path.cont[newReq.path.top - 1 - i];
	newReq.path.cont[newReq.path.top - 1 - i] = tmp;
    }

    // eval the impact of the addition
    if ((linkID = DBgetLinkID(dataBase, src, dst)) >= 0 && 
	DBevalLSOnSetup(dataBase, src, dst, &newLS, lsList[linkID], &newReq) >= 0)
    {
	for (i=0; i<NB_OA; ++i) 
	{
	    bw_tot += lsp->bw[i];

	    bw_before[i] = 0;
	    bw_after[i] = 0;
	    for (j=0; j<NB_PREEMPTION; ++j)
	    {
		bw_after[i] += newLS.rbw[i][j];
		bw_before[i] += lsList[linkID]->rbw[i][j];
	    }
	    bw_tot_bef += bw_before[i];
	    bw_tot_aft += bw_after[i];
	}

	// printf("%ld-%ld : %f\n", src, dst, bw_after - bw_before);

    }
    else
    {
	addError(WARNING,"Error while computing new link state in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    lspRequestEnd(&newReq);

    DBlinkStateEnd(&newLS);

    // Capacity constrain ....
    for (i=0; i<NB_OA; i++)
	if (bw_after[i] > newLS.cap[i]) {
	    return -1;
	}

    inc = bw_tot_aft - bw_tot_bef;
    if (inc < 0) {
	inc = 0;
    }

    // ------- Hop count ---------
    // ---------------------------

    // if we merge with the primary we have to account for the remaining of the path
    
    for (i=0; i<lsp->primPath.top; i++) {
	if (lsp->primPath.cont[i] == dst) {
	    break;
	}
    }

    return (alpha * inc) + (beta * (lsp->primPath.top + 1 - i) * bw_tot); 
}



/**
 * Backup LSP computation function
 * @param dataBase the general database containing topology
 * @param reqList the structure where computed backup lsp(s) will be stored 
 * @param lspID the ID of the primary LSP for which local or global backup(s) are computed
 * @param type the type of the backup LSP(s) to be computed.  If type is LOCAL_BACK, local backups will be computed, otherwise if type is GLOBAL_BACK, one global backup will be computed. 
 */

int computeBackup(DataBase* dataBase, LSPRequestList* reqList, long lspID, BackupType type)
{
    long i,j,pNodeIndex,start,pNode;
    DBLinkState *ls, *oldLS;
    DBLinkState** lsList;
    CPDijkNode *dn=NULL; 
    CPDijkNode** nodeList;
    bool reachPrimary;
    CPPrioQueue toBeTreated;
    LongList* neigh;
    LongList forbiddenLinks;
    DBLabelSwitchedPath* lsp;
    LSPRequest* req=NULL;
    double newVal;
    int pType=0;
    int src, dst;

    enum {NODE_FAILURE, LINK_FAILURE};

#if defined LINUX && defined TIMING && defined TIME3
    struct timezone tz;
    struct timeval  t1,t2;
#endif

    if (dataBase == NULL || reqList == NULL)
    {
	addError(CRITICAL,"Wrong argument in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (lspID < 0 || ((lsp = DBgetLSP(dataBase, lspID)) == NULL))
    {
	addError(CRITICAL,"Cannot find lsp ID %ld in %s at line %d",
		 lspID,__FILE__,__LINE__);
	return -1;
    }

    if ((lsList = calloc(dataBase->linkSrcVec.top, sizeof(DBLinkState*))) == NULL)
    {
	addError(CRITICAL,"Cannot allocate a required structure in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((nodeList = calloc(dataBase->nodeVec.top, sizeof(CPDijkNode*))) == NULL)
    {
	addError(CRITICAL,"Cannot allocate a required structure in %s at line %d",
		 __FILE__,__LINE__);
	free (lsList);
	return -1;
    }

    // duplicate all the link-states
    for (i=0; i<dataBase->linkSrcVec.top; i++)
    {
	src = DBgetLinkSrc(dataBase, i);
	dst = DBgetLinkDst(dataBase, i);

	if (src != -1 && dst != -1)
	{
	    if ((oldLS = DBgetLinkState(dataBase, src, dst)) == NULL)
	    {
		addError(WARNING,"Oups there should be a link-state here in %s at line %d",
			 __FILE__,__LINE__);
		continue;
	    }
	    
	    if ((ls = DBlinkStateNew()) == NULL)
	    {
		addError(CRITICAL,"Cannot duplicate all the link-states in %s at line %d",
			 __FILE__,__LINE__);
		continue;
	    }

	    if (DBlinkStateCopy(ls, oldLS) < 0)
	    {
		addError(CRITICAL,"Something went wrong while copying in %s at line %d",
		 __FILE__,__LINE__);
		DBlinkStateDestroy(ls);
		continue;
	    }

	    lsList[i] = ls;
	}
	else 
	{
	    addError(INFO,"Warning there is no link numbered %ld : src = %ld, dst = %ld .... in %s at line %ld", i, src, dst, 
		     __FILE__, __LINE__);
	}
    }

    // create the Dijk Nodes ... used for the computation
    for (i=0; i<dataBase->nodeVec.top; i++)
    {
	if (dataBase->nodeVec.cont[i] != NULL)
	{
	    if ((dn = calloc(1,sizeof(CPDijkNode))) == NULL)
	    {
		addError(CRITICAL,"Cannot create the Dijk nodes in %s at line %d",
			 __FILE__,__LINE__);
		continue;
	    }

	    dn->node = dataBase->nodeVec.cont[i];
	    dn->val = -1;
	    dn->marked = FALSE;

	    nodeList[i] = dn;
	}
    }

    /* printf("Primary path :");
    for (i=0; i<lsp->path.top - 1; ++i)
    {
	printf("%ld - ", lsp->path.cont[i]);
    }
    printf("%ld\n", lsp->path.cont[i]);*/
    
    // now start the calculation ....
    
#if defined LINUX && defined TIMING && defined TIME3
    gettimeofday(&t1, &tz);
#endif

    // init a PrioQueue
    CPinitPQ(&toBeTreated);

    // init the forbiddenLinks;
    longListInit(&forbiddenLinks, -1);

    if (type == LOCAL)
    {
	// init the list of request to return;
	lspRequestListResize(reqList, lsp->path.top-1);
	for (i=0; i<lsp->path.top-1; ++i)
	{
	    req = lspRequestListGet(reqList, i);

	    req->primID = lsp->id;
	    req->type = LOCAL_BACK;
	    if (lsp->precedence + 1 < NB_PREEMPTION)
		req->precedence = lsp->precedence + 1;
	    else
		req->precedence = lsp->precedence;
	    memmove(&(req->bw),&(lsp->bw), NB_OA * sizeof(double));
	}
	
	// for (pNodeIndex=1; pNodeIndex<req->path.top; ++pNodeIndex) // start at 1 because we cannot protect node 0
	for (pNodeIndex=lsp->path.top - 1; pNodeIndex>0; --pNodeIndex)
	{
	    pNode = lsp->path.cont[pNodeIndex];
	    start = lsp->path.cont[pNodeIndex - 1]; 
	    
	    pType = NODE_FAILURE;
	    
	    // mark the forbidden links
	    forbiddenLinks.top = 0;
	    
	    if (pType == NODE_FAILURE)
	    {
		neigh = DBgetNodeInNeighb(dataBase, pNode);
		if (neigh == NULL)
		{
		    addError(CRITICAL,"The protected node must have some neighbour in %s at line %d",
			     __FILE__,__LINE__);
		return -1;
		}
		
		for (i=0; i<neigh->top; ++i)
		{
		    longListPushBack(&forbiddenLinks, DBgetLinkID(dataBase, neigh->cont[i], pNode));
		}
	    }
	    else if (pType == LINK_FAILURE)
	    {
		longListPushBack(&forbiddenLinks, DBgetLinkID(dataBase, start, pNode));
	    }
	    
	    // clear the PQ;
	    while (CPpopTop(&toBeTreated) != NULL);
	    
	    // clear all the marks
	    for (i=0; i<dataBase->nodeVec.top; i++)
	    {
		nodeList[i]->marked = FALSE;
		nodeList[i]->val = -1;
		nodeList[i]->from = NULL;
	    }
	    
	    // push the first node on the PQ
	    CPinsertPQ(&toBeTreated, nodeList[start], (nodeList[start]->val=0));
	
	    reachPrimary = FALSE;
	    while (reachPrimary == FALSE)
	    {
		if ((dn = CPpopTop(&toBeTreated)) == NULL)
		{
		    // Oups ... impossible to reach the primary
		    // if we are in node protection mode, switch back to link protection
		    if (pType == NODE_FAILURE)
		    {
			pType = LINK_FAILURE;
			
			// mark the forbidden links
			forbiddenLinks.top = 0;
			longListPushBack(&forbiddenLinks, DBgetLinkID(dataBase, start, pNode));
			
			// clear all marked nodes
			for (i=0; i<dataBase->nodeVec.top; i++)
			{
			    nodeList[i]->marked = FALSE;
			    nodeList[i]->val = -1;
			    nodeList[i]->from = NULL;
			}
			
			// push the first node on the PQ
			CPinsertPQ(&toBeTreated, nodeList[start], (nodeList[start]->val=0));
			
			// re-enter the loop
			continue;
		    }
		    else
		    {
			break;
		    }
		}
		
		// as we don't remove marked node immediatelly we may encounter one now so we skip it
		if (dn->marked == TRUE)
		    continue;
		
		// mark the node
		dn->marked = TRUE;
		
		// check the stop condition
		for (i=pNodeIndex; i<lsp->path.top; ++i)
		    if (lsp->path.cont[i] == dn->node->id)
		    {
			reachPrimary = TRUE;
			break;
		    }
		
		// we have finished ... leave the while loop
		if (reachPrimary == TRUE)
		    break;
		
		// find the neighbours
		neigh = DBgetNodeOutNeighb(dataBase, dn->node->id);
		
		if (neigh != NULL)
		{
		    for (i=0; i<neigh->top; ++i)
		    {
			int id;
			double cost;

			// check if the node is not already marked
			if (nodeList[neigh->cont[i]]->marked == TRUE)
			    continue;
			
			// check if the link is valid
			id = DBgetLinkID(dataBase, dn->node->id, neigh->cont[i]);
			for (j=0; j<forbiddenLinks.top; ++j)
			    if (forbiddenLinks.cont[j] == id)
				break;
			
			if (j != forbiddenLinks.top)
			    continue;
			
			// ok now update the node ...
			cost = computeCost(dataBase, lsList, dn->node->id, neigh->cont[i], dn, lsp, type);
			if (cost >= 0) {
			    newVal = dn->val + cost;
			
			    if (nodeList[neigh->cont[i]]->val == -1 || (newVal > 0 && newVal < nodeList[neigh->cont[i]]->val))
			    {
				nodeList[neigh->cont[i]]->val = newVal;
				nodeList[neigh->cont[i]]->from = dn;
				CPinsertPQ(&toBeTreated, nodeList[neigh->cont[i]], newVal);
			    }
			}
		    }
		}
	    }
	    
	    if (reachPrimary == TRUE)
	    {
		req = lspRequestListGet(reqList, pNodeIndex-1);

		// clear the previous link state modification
		for (i=0; i<req->path.top - 1; i++)
		{
		    int lnk = DBgetLinkID(dataBase, req->path.cont[i], req->path.cont[i+1]);
		    DBevalLSOnRemove(dataBase, req->path.cont[i], req->path.cont[i+1],
				     lsList[lnk], lsList[lnk], req);
		}
		
		// clear the old path ...
		req->path.top = 0;
		
		// ok we found a path ...
		//printf("Cost = %f, Path = ", dn->val);

		while (dn != NULL && dn->from != NULL)
		{
		    longListPushBack(&(req->path), dn->node->id);
		    dn = dn->from;
		}
		longListPushBack(&(req->path), dn->node->id);
		
		// revert the path
		for (i=0; i<(req->path.top + 1)/2; i++)
		{
		    int tmp = req->path.cont[i];
		    req->path.cont[i] = req->path.cont[req->path.top - 1 - i];
		    req->path.cont[req->path.top - 1 - i] = tmp;
		}
		
		for (i=0; i<req->path.top - 1; i++)
		{
		    int lnk = DBgetLinkID(dataBase, req->path.cont[i], req->path.cont[i+1]);
		    DBevalLSOnSetup(dataBase, req->path.cont[i], req->path.cont[i+1],
				    lsList[lnk], lsList[lnk], req);
		}

		
		/*for (i=0; i<req->path.top - 1; i++)
		{
		    printf("%ld - ", req->path.cont[i]);
		}
		printf("%ld\n", req->path.cont[i]);*/
	    }
	    else 
	    {
		// oups we have to reject the request ...
	    fprintf(stderr,"No local link disjoint protection path found\n");
	    return -1;
		
	    }
	    
	}
    }
    else if (type == GLOBAL)
    {
	
	// init the list of request to return;
	lspRequestListResize(reqList, 1); // should not be required !
	req = lspRequestListGet(reqList, 0);
	
	req->primID = lsp->id;
	req->type = GLOBAL_BACK;
	if (lsp->precedence + 1 <NB_PREEMPTION)
	    req->precedence = lsp->precedence + 1;
	else
	    req->precedence = lsp->precedence;
	memmove(&(req->bw),&(lsp->bw), NB_OA * sizeof(double));
	
	start = lsp->path.cont[0]; 
	    
	pType = NODE_FAILURE;
	    
	// mark the forbidden links
	forbiddenLinks.top = 0;
	    
	if (pType == NODE_FAILURE)
	{
	    // don't remove first and last node !!!
	    for (i=1; i<lsp->path.top-1; i++)
	    {
		neigh = DBgetNodeInNeighb(dataBase, lsp->path.cont[i]);
		if (neigh == NULL)
		{
		    addError(CRITICAL,"The protected node must have some neighbour in %s at line %d",
			     __FILE__,__LINE__);
		    return -1;
		}
		
		for (j=0; j<neigh->top; ++j)
		{
		    longListPushBack(&forbiddenLinks, DBgetLinkID(dataBase, neigh->cont[j], lsp->path.cont[i]));
		}
	    }

	    // last link in the path must be removed !!!
	    longListPushBack(&forbiddenLinks, DBgetLinkID(dataBase, lsp->path.cont[lsp->path.top-2], lsp->path.cont[lsp->path.top-1]));

	}
	else if (pType == LINK_FAILURE)
	{
	    for (i=1; i<lsp->path.top; i++)
	    {
		longListPushBack(&forbiddenLinks, DBgetLinkID(dataBase, lsp->path.cont[i-1], lsp->path.cont[i]));
	    }
	}
	    
	// clear the PQ;
	while (CPpopTop(&toBeTreated) != NULL);
	    
	// clear all the marks
	for (i=0; i<dataBase->nodeVec.top; i++)
	{
	    nodeList[i]->marked = FALSE;
	    nodeList[i]->val = -1;
	    nodeList[i]->from = NULL;
	}
	    
	// push the first node on the PQ
	CPinsertPQ(&toBeTreated, nodeList[start], (nodeList[start]->val=0));
	
	reachPrimary = FALSE;
	while (reachPrimary == FALSE)
	{
	    if ((dn = CPpopTop(&toBeTreated)) == NULL)
	    {
		// Oups ... impossible to reach the primary
		// if we are in node protection mode, switch back to link protection
		if (pType == NODE_FAILURE)
		{
		    printf("Switching from node to link protection for lsp number: %ld\n",lsp->id);

		    pType = LINK_FAILURE;
			
		    // mark the forbidden links
		    forbiddenLinks.top = 0;
		    for (i=1; i<lsp->path.top; i++)
		    {
			longListPushBack(&forbiddenLinks, DBgetLinkID(dataBase, lsp->path.cont[i-1], lsp->path.cont[i]));
		    }
			
		    // clear all marked nodes
		    for (i=0; i<dataBase->nodeVec.top; i++)
		    {
			nodeList[i]->marked = FALSE;
			nodeList[i]->val = -1;
			nodeList[i]->from = NULL;
		    }
			
		    // push the first node on the PQ
		    CPinsertPQ(&toBeTreated, nodeList[start], (nodeList[start]->val=0));
			
		    // re-enter the loop
		    continue;
		}
		else
		{
		    printf("Oups ... no path found ...\n");
		    break;
		}
	    }
		
	    // as we don't remove marked node immediatelly we may encounter one now so we skip it
	    if (dn->marked == TRUE)
		continue;
		
	    // mark the node
	    dn->marked = TRUE;
		
	    // check the stop condition
	    if (lsp->path.cont[lsp->path.top-1] == dn->node->id)
		reachPrimary = TRUE;
		
	    // we have finished ... leave the while loop
	    if (reachPrimary == TRUE)
		break;
		
	    // find the neighbours
	    neigh = DBgetNodeOutNeighb(dataBase, dn->node->id);
		
	    if (neigh != NULL)
	    {
		for (i=0; i<neigh->top; ++i)
		{
		    int id;
		    double cost;
		    
		    // check if the node is not already marked
		    if (nodeList[neigh->cont[i]]->marked == TRUE)
			continue;
		    
		    // check if the link is valid
		    id = DBgetLinkID(dataBase, dn->node->id, neigh->cont[i]);
		    for (j=0; j<forbiddenLinks.top; ++j)
			if (forbiddenLinks.cont[j] == id)
			    break;
		    
		    if (j != forbiddenLinks.top)
			continue;
		    
		    /*   // ok now update the node ...
			 newVal = dn->val + computeCost(dataBase, lsList, dn->node->id, neigh->cont[i], dn, lsp, type);
			 
			 if (nodeList[neigh->cont[i]]->val == -1 || (newVal > 0 && newVal < nodeList[neigh->cont[i]]->val))
			 {
			 nodeList[neigh->cont[i]]->val = newVal;
			 nodeList[neigh->cont[i]]->from = dn;
			 CPinsertPQ(&toBeTreated, nodeList[neigh->cont[i]], newVal);
			 }*/
// ok now update the node ...
		    cost = computeCost(dataBase, lsList, dn->node->id, neigh->cont[i], dn, lsp, type);
		    if (cost >= 0) {
			newVal = dn->val + cost;
			
			if (nodeList[neigh->cont[i]]->val == -1 || (newVal > 0 && newVal < nodeList[neigh->cont[i]]->val))
			{
			    nodeList[neigh->cont[i]]->val = newVal;
			    nodeList[neigh->cont[i]]->from = dn;
			    CPinsertPQ(&toBeTreated, nodeList[neigh->cont[i]], newVal);
			}
		    }

		}
	    }
	}
	
	if (reachPrimary == TRUE)
	{
	    req = lspRequestListGet(reqList, 0);

	    // clear the previous link state modification
	    for (i=0; i<req->path.top - 1; i++)
	    {
		int lnk = DBgetLinkID(dataBase, req->path.cont[i], req->path.cont[i+1]);
		DBevalLSOnRemove(dataBase, req->path.cont[i], req->path.cont[i+1],
				 lsList[lnk], lsList[lnk], req);
	    }
		
	    // clear the old path ...
	    req->path.top = 0;
	    
	    // ok we found a path ...
	    //printf("Cost = %f, Path = ", dn->val);

	    while (dn != NULL && dn->from != NULL)
	    {
		longListPushBack(&(req->path), dn->node->id);
		dn = dn->from;
	    }
	    longListPushBack(&(req->path), dn->node->id);
	    
	    // revert the path
	    for (i=0; i<(req->path.top + 1)/2; i++)
	    {
		int tmp = req->path.cont[i];
		req->path.cont[i] = req->path.cont[req->path.top - 1 - i];
		req->path.cont[req->path.top - 1 - i] = tmp;
	    }
	    
	    for (i=0; i<req->path.top - 1; i++)
	    {
		int lnk = DBgetLinkID(dataBase, req->path.cont[i], req->path.cont[i+1]);
		DBevalLSOnSetup(dataBase, req->path.cont[i], req->path.cont[i+1],
				lsList[lnk], lsList[lnk], req);
	    }

		
	    /*for (i=0; i<req->path.top - 1; i++)
	    {
		printf("%ld - ", req->path.cont[i]);
	    }
	    printf("%ld\n", req->path.cont[i]);*/
	}
	else 
	{
	    // oups we have to reject the request ...
	    fprintf(stderr,"No end to end link disjoint protection path found\n");
	    return -1;
		
	}	    
    }
    else if (type == NONE)
    {
	addError(INFO,"Oups no backup were requested in %s at line %d",
			 __FILE__,__LINE__);
    }
    else
    {
	// error
	addError(WARNING,"Unknown backup type in %s at line %d",
		 __FILE__,__LINE__);
    }
	
#if defined LINUX && defined TIMING && defined TIME3
    gettimeofday(&t2, &tz);
    fprintf(stderr, "Time for calculation of backups paths : %f ms\n", (t2.tv_sec - t1.tv_sec) * 1000 + 
	    (t2.tv_usec - t1.tv_usec) / 1000.0);
#endif

    ///////////////////////
    // Clean Up
    //////////////////////

    // clear the PrioQueue
    CPendPQ(&toBeTreated);
    
    // clear the forbiddenLinks
    longListEnd(&forbiddenLinks);

    // clear the local copy ...
    for (i=0; i<dataBase->linkSrcVec.top; i++)
    {
	if (lsList[i] != NULL)
	{
	    if (DBlinkStateDestroy(lsList[i]) < 0)
	    {
		addError(CRITICAL,"Something went wrong while clearing a structure in %s at line %d",
			 __FILE__,__LINE__);
	    }
	}
    }

    if (lsList != NULL)
	free(lsList);

    // free the dijkNodes
    for (i=0; i<dataBase->nodeVec.top; i++)
    {
	if (nodeList[i] != NULL)
	{
	    free(nodeList[i]);
	}
    }

    if (nodeList)
	free(nodeList);

    return 0;
}

