#include "database_util.h"
#include "database_st.h"
#include "database_api.h"

#include <stdio.h>
#include <string.h>


/******************************************************************
 *    Structures and substructures of the database and handlers   * 
 ******************************************************************/


//Internal representation of a node in the DB
//--------------------------------------------

///return a newly (dynamically) allocated DBNode object. 
DBNode* DBnodeNew() 
{
    DBNode* ptr=NULL;

    if ((ptr = calloc(1,sizeof(DBNode))) == NULL) 
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (longListInit(&(ptr->inNeighb),-1) < 0)
    {
	addError(CRITICAL,"Unable to initialize the incoming neighbour list in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (longListInit(&(ptr->outNeighb),-1) < 0)
    {
	addError(CRITICAL,"Unable to initialize the outgoing neighbour list in %s at line %d",
		 __FILE__,__LINE__);
	longListEnd(&(ptr->inNeighb));
	free(ptr);
	return NULL;
    }

    return ptr;
}

///initialize a DBNode object allready allocated somewhere else.
int DBnodeInit(DBNode* node) 
{
    if (node == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (longListInit(&(node->inNeighb),-1) < 0) 
    {
	addError(CRITICAL,"Unable to initialize the incoming neighbour list in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
 
    if (longListInit(&(node->outNeighb),-1) < 0) 
    {
	addError(CRITICAL,"Unable to initialize the outgoing neighbour list in %s at line %d",
		 __FILE__,__LINE__);
	longListEnd(&(node->inNeighb));
	return -1;
    }

    return 0;
} 

///clear function to free the ressources of a Link object allocated on the heap.
int DBnodeDestroy(DBNode* node) 
{
    if (node == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    longListEnd(&(node->inNeighb));
    longListEnd(&(node->outNeighb));
    free(node);
   
    return 0;
}

///clear function to free the ressources of a Node object allocated on the stack. 
int DBnodeEnd(DBNode* node) 
{
    if (node == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    longListEnd(&(node->inNeighb));
    longListEnd(&(node->outNeighb));
   
    return 0;
}


void DBprintNode(DBNode* node) 
{
    long i;

    if (node == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return;
    }
    
    printf("Incoming neighboors : \n");

    for (i=0; i<node->inNeighb.top; i++)
    {
	printf("%ld ", node->inNeighb.cont[i]);
    }

    printf("\nOutgoing neighboors : \n");

    for (i=0; i<node->outNeighb.top; i++)
    {
	printf("%ld ", node->outNeighb.cont[i]);
    }

    printf("\n");
}



//Internal representation of a link in the DB
//--------------------------------------------

///return a newly (dynamically) allocated Link object. 
DBLink* DBlinkNew()
{
    DBLink* ptr=NULL;
    
    if ((ptr = calloc(1,sizeof(DBLink))) == NULL) 
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (DBlinkStateInit(&(ptr->state)) == -1) 
    {
	addError(CRITICAL,"Error while initializing LinkState in %s at line %d",
		 __FILE__,__LINE__);
	free(ptr);
	return NULL;
    }

    if (DBlspListInit(&(ptr->lspList),-1) < 0)
    {
	addError(CRITICAL,"Error while initializing LinkState in %s at line %d",
		 __FILE__,__LINE__);
	DBlinkStateEnd(&(ptr->state));
	free(ptr);
	return NULL;
    }

    return ptr;
}

///initialize a Link object allready allocated somewhere else.
int DBlinkInit(DBLink* link)
{
    if (link == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (DBlinkStateInit(&(link->state)) == -1) 
    {
	addError(CRITICAL,"Error while initializing LinkState in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (DBlspListInit(&(link->lspList),-1) < 0)
    {
	addError(CRITICAL,"Error while initializing LinkState in %s at line %d",
		 __FILE__,__LINE__);
	DBlinkStateEnd(&(link->state));
	return -1;
    }

    return 0;
}

///clear function to free the ressources of a Link object allocated on the heap.
int DBlinkDestroy(DBLink* link)
{
    if (link == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    DBlinkStateEnd(&link->state);
    DBlspListEnd(&(link->lspList));
    free(link);

    return 0;
}

///clear function to free the ressources of a Link object allocated on the stack. 
int DBlinkEnd(DBLink* link)
{
    if (link == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    DBlinkStateEnd(&link->state);
    DBlspListEnd(&(link->lspList));

    return 0;
}


void DBprintLink(DBLink *link)
{
    long i,oa;
    double ptot,rtot;
    DBLabelSwitchedPath* lsp=NULL;

    if (link == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return;
    }

    printf("\tList of LSPs\n");
    printf("\t------------\n");

    for (i=0; i<link->lspList.top; ++i)
    {
	lsp = link->lspList.cont[i];
	printf("%ld ", lsp->id);
    }

    printf("\n\n");
    
    printf("\tLink-state\n");
    printf("\t----------\n");
	
    for (oa=0; oa<NB_OA; ++oa)
    {
	ptot = 0;
	rtot = 0;

	printf("\tCapacity[%ld] = %f\n", oa, link->state.cap[oa]);

	for (i=0; i<NB_PREEMPTION; ++i)
	{
	    ptot += link->state.pbw[oa][i];
	    rtot += link->state.rbw[oa][i];
	}

	printf("\tpbw[%ld] = %f\n", oa, ptot);
	printf("\trbw[%ld] = %f\n", oa, rtot);
    }

    printf("\n\n");

}



// Collection of all nodes (pointers) representation
//---------------------------------------------------

DBNodeVec* DBnodeVecNew(long size)
{
    DBNodeVec *vec=NULL;  
    void* ptr=NULL;
 
    if ((vec = calloc(1,sizeof(DBNodeVec))) == NULL) 
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (size == -1)
	size = NODEVEC_INITSIZE;

    if ((ptr = calloc(size,sizeof(DBNode*))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	free(vec);
	return NULL;
    }
    
    vec->size = size;
    vec->top = 0;
    vec->cont = ptr;

    return vec;
}


int DBnodeVecInit(DBNodeVec *vec,long size)
{
    void* ptr=NULL;
    
    if (vec == NULL)
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (size == -1)
	size = NODEVEC_INITSIZE;

    if ((ptr = calloc(size,sizeof(DBNode*))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    vec->size = size;
    vec->top = 0;    
    vec->cont = ptr;

    return 0;   
}

int DBnodeVecDestroy(DBNodeVec *vec)
{
    int i;

    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    for (i=0; i<vec->size; i++)
    {
	if (vec->cont[i]!=NULL)
	{
	    DBnodeDestroy(vec->cont[i]);
	}
    }
    
    free(vec->cont);
    free(vec);

    return 0;
}

int DBnodeVecEnd(DBNodeVec *vec)
{
    int i;

    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    for (i=0; i<vec->size; i++)
    {
	if (vec->cont[i]!=NULL)
	{
	    DBnodeDestroy(vec->cont[i]);
	}
    }

    free(vec->cont);
    vec->cont = NULL;
    vec->size = 0;
    vec->top = 0;

    return 0;
}

int DBnodeVecResize(DBNodeVec *vec,long size)
{
    void *ptr=NULL;
    int i;

    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (size < vec->size)
    {
	for (i=size;i<vec->size;i++)
	{
	    if (vec->cont[i]!=NULL)
	    {
		DBnodeDestroy(vec->cont[i]);
		vec->cont[i]=NULL;
	    }
	}

	if (size < vec->top)
	{
	    vec->top = size;
	    while (vec->cont[vec->top-1] == NULL)
		vec->top--;
	}

    }

    

    if ((ptr = realloc(vec->cont, size * sizeof(DBNode*))) == NULL) 
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (size > vec->size)
    {
	memset(ptr + (vec->size * sizeof(DBNode*)), 0, (size-vec->size) * sizeof(DBNode*));
    }

    vec->size=size;
    vec->cont=ptr;

    return 0;
}

int DBnodeVecSet(DBNodeVec *vec,DBNode *node,long id)
{
    if (vec == NULL || vec->cont == NULL || node == NULL || id <0) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (id >= vec->size)
    {
	if (DBnodeVecResize(vec,max(2*vec->size,id+1))<0)
	{
	    addError(CRITICAL,"Unable to resize node vector prior to insertion in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;  
	}
    }

    if (vec->cont[id] != NULL)
    {
	addError(CRITICAL,"Trying to add a node with a reserved ID in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    vec->cont[id]=node;
    vec->top = max(vec->top, id+1);

    return 0;
}

int DBnodeVecRemove(DBNodeVec *vec,long id)
{
    if (vec == NULL || vec->cont == NULL ||
	id <0 || id >= vec->size || vec->cont[id] == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    DBnodeDestroy(vec->cont[id]);
    vec->cont[id]=NULL;

    while (vec->cont[vec->top-1] == NULL)
	vec->top--;
    
    return 0;
}



// Collection of all LSPs (pointers) representation
//--------------------------------------------------

DBLSPVec* DBlspVecNew(long size)
{
    DBLSPVec *vec=NULL;  
    void* ptr=NULL;
 
    if ((vec = calloc(1,sizeof(DBLSPVec))) == NULL) 
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (size == -1)
	size = LSPVEC_INITSIZE;

    if ((ptr = calloc(size,sizeof(DBLabelSwitchedPath*))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	free(vec);
	return NULL;
    }
    
    vec->size = size;
    vec->cont = ptr;

    return vec;   
}

int DBlspVecInit(DBLSPVec *vec,long size)
{
    void* ptr=NULL;
    
    if (vec == NULL)
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (size == -1)
	size = LSPVEC_INITSIZE;

    if ((ptr = calloc(size,sizeof(DBLabelSwitchedPath*))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    vec->size = size;
    vec->cont = ptr;

    return 0;   
}

int DBlspVecDestroy(DBLSPVec *vec)
{
    int i;

    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    for (i=0; i<vec->size; i++)
    {
	if (vec->cont[i]!=NULL)
	{
	    DBlspDestroy(vec->cont[i]);
	}
    }
    
    free(vec->cont);
    free(vec);

    return 0;
}

int DBlspVecEnd(DBLSPVec *vec)
{
    int i;
    
    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    for (i=0; i<vec->size; i++)
    {
	if (vec->cont[i]!=NULL)
	{
	    DBlspDestroy(vec->cont[i]);
	}
    }
    
    free(vec->cont);
    vec->cont = NULL;
    vec->size = 0;

    return 0;
}

int DBlspVecResize(DBLSPVec *vec,long size)
{
    void *ptr=NULL;
    int i;

    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (size < vec->size)
    {
	for (i=size;i<vec->size;i++)
	{
	    if (vec->cont[i]!=NULL)
	    {
		DBlspDestroy(vec->cont[i]);
		vec->cont[i]=NULL;
	    }
	}
    }

    if ((ptr = realloc(vec->cont, size * sizeof(DBLabelSwitchedPath*))) == NULL) 
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (size > vec->size)
    {
	memset(ptr + (vec->size * sizeof(DBLabelSwitchedPath*)), 0, (size-vec->size) * sizeof(DBLabelSwitchedPath*));
    }

    vec->size=size;
    vec->cont=ptr;

    return 0;
}

int DBlspVecSet(DBLSPVec *vec,DBLabelSwitchedPath *lsp,long id)
{
    if (vec == NULL || vec->cont == NULL || id <0) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (id >= vec->size)
    {
	if (DBlspVecResize(vec,max(2*vec->size,id+1))<0)
	{
	    addError(CRITICAL,"Unable to resize LSP vector prior to insertion in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;  
	}
    }

    if (vec->cont[id] != NULL)
    {
	addError(CRITICAL,"Trying to add an LSP with a reserved ID in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    vec->cont[id]=lsp;

    return 0;
}

int DBlspVecRemove(DBLSPVec *vec,long id)
{
    if (vec == NULL || vec->cont == NULL ||
	id <0 || id >= vec->size || vec->cont[id] == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    vec->cont[id]=NULL;

    return 0;
}




// Collection of all links (pointers) representation
//----------------------------------------------------

DBLinkTab* DBlinkTabNew(long size)
{
    DBLinkTab *tab=NULL;  
    DBLink ***ptr=NULL;
    int i;

    if ((tab = calloc(1,sizeof(DBLinkTab))) == NULL) 
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (size == -1)
	size = LINKTAB_INITSIZE;

    if ((ptr=(calloc(size,sizeof(DBLink**))))==NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	free(tab);
	return NULL;
    }
    else
    {
	for (i=0;i<size;i++)
	{
	    if ((ptr[i]=(calloc(size, sizeof(DBLink*))))==NULL)
	    { 
		addError(CRITICAL,"Critical lack of memory in %s at line %d",
			 __FILE__,__LINE__);
		for (i=i-1;i>=0;i--)
		    free(ptr[i]);
		free(ptr);
		free(tab);
		return NULL;
	    }
	}
    }
    
    tab->size=size;
    tab->cont=ptr;

    return tab;
}

int DBlinkTabInit(DBLinkTab *tab,long size)
{
    DBLink ***ptr=NULL;
    int i;

    if (tab == NULL)
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (size == -1)
	size = LINKTAB_INITSIZE;
    
    if ((ptr=(calloc(size,sizeof(DBLink**))))==NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    else
    {
	for (i=0;i<size;i++)
	{
	    if ((ptr[i]=(calloc(size, sizeof(DBLink*))))==NULL)
	    { 
		addError(CRITICAL,"Critical lack of memory in %s at line %d",
			 __FILE__,__LINE__);
		for (i=i-1;i>=0;i--)
		    free(ptr[i]);
		free(ptr);
		return -1;
	    }
	}
    }
    
    tab->size=size;
    tab->cont=ptr;

    return 0;
}

int DBlinkTabDestroy(DBLinkTab *tab)
{
    int i,j;
    
    if (tab == NULL || tab->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    for (i=0;i<tab->size;i++)
    {
	for (j=0; j<tab->size; j++)
	{
	    if (tab->cont[i][j] != NULL) 
	    {
		DBlinkDestroy(tab->cont[i][j]);
	    }
	}
	free(tab->cont[i]);
    }
    free(tab->cont);
    free(tab);

    return 0;
}

int DBlinkTabEnd(DBLinkTab *tab)
{
    int i,j;
    
    if (tab == NULL || tab->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    for (i=0;i<tab->size;i++)
    {
	for (j=0; j<tab->size; j++)
	{
	    if (tab->cont[i][j] != NULL) 
	    {
		DBlinkDestroy(tab->cont[i][j]);
	    }
	}
	free(tab->cont[i]);
    }

    free(tab->cont);
    tab->cont = NULL;
    tab->size = 0;

    return 0;
}


int DBlinkTabResize(DBLinkTab *tab, long size)
{
    DBLink*** ptr=NULL;
    DBLink** ptr2=NULL;
    int i,j;

    if (tab == NULL || tab->cont == NULL)
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (size < tab->size)
    {
	for (i=size;i<tab->size;i++)
	{
	    for (j=size; j<tab->size; j++)
	    {
		if (tab->cont[i][j] != NULL) 
		{
		    DBlinkDestroy(tab->cont[i][j]);
		    tab->cont[i][j]=NULL;
		}
	    }
	}
	free(tab->cont[i]);
    }

    if ((ptr = realloc(tab->cont,size * sizeof(DBLink**))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    else
    {
	tab->cont = ptr;

	for (i=0;i<min(tab->size,size);i++)
	{
	    if ((ptr2 = realloc(ptr[i], size * sizeof(DBLink*)))==NULL)
	    { 
		addError(CRITICAL,"Critical lack of memory in %s at line %d",
			 __FILE__,__LINE__);
		tab->size=min(tab->size,size);
		return -1;
	    }
	    
	    ptr[i] = ptr2;

	    if (size > tab->size)
	    {
		memset(ptr2 + tab->size, 0, (size-tab->size) * sizeof(DBLink*));
	    }
	}
	
	if (size > tab->size)
	{
	    for (i=tab->size;i<size;i++)
	    {
		if ((ptr[i] = calloc(size, sizeof(DBLink*)))==NULL)
		{ 
		    addError(CRITICAL,"Critical lack of memory in %s at line %d",
			     __FILE__,__LINE__);
		    tab->size=i;
		    return -1;
		}
	    }
	}
    }

    tab->size=size;

    return 0;
}

int DBlinkTabSet(DBLinkTab *tab,DBLink *lnk,long src,long dst)
{
    long resize;

    if (tab == NULL || tab->cont == NULL || src <0 || dst<0) 
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    resize=max(src,dst)+1;

    if (resize > tab->size)
    {
	if (DBlinkTabResize(tab,max(2*tab->size,resize))<0)
	{
	    addError(CRITICAL,"Unable to resize link table prior to insertion in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;  
	}
    }

    tab->cont[src][dst]=lnk;

    return 0;
}

int DBlinkTabRemove(DBLinkTab *tab,long src,long dst)
{
    if (tab == NULL || tab->cont == NULL ||
	src <0 || dst<0 || src >= tab->size || dst >= tab->size ||
	tab->cont[src][dst] == NULL) 
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    DBlinkDestroy(tab->cont[src][dst]);
    tab->cont[src][dst]=NULL;
    
    return 0;
}


